/* 
 * File:   Image.cpp
 * Author: tg
 * 
 * Created on December 11, 2013, 11:03 PM
 */

#include "Image.h"
#include "defines.h"
#include "myExceptions.h"

#include <forward_list>
#include <stdexcept>
#include <iostream>
#include <cmath>
#include <algorithm>

template <typename T>Image<T>::Image(const std::vector<T>& data, int width, int height)
: data_(data.begin(), data.end()), width_(width), height_(height), mask_(data.size(), false) {
    typename std::vector<T>::const_iterator it;
    // setting zero elements to mask
    for (it = data_.begin(); it != data_.end(); ++it) {
        if (*it == 0) {
            const size_t idx (it-data_.begin());
            mask_[idx] = true;
        }
    }
}

template <>Image<unsigned char>::Image(const std::vector<unsigned char>& data, int width, int height)
: width_(width), height_(height), mask_(data.size(), false) {
    data_.reserve(data.size());
    for (int h = 0; h < height_; ++h) {
        std::vector<unsigned char>::const_iterator it;

        it = data.begin() + h*width_;
        for (int w = 0; w < width_; ++w) {
            unsigned short val = *it;

            if (val == 0) {
                data_.push_back(0);
            } else {
                val = round(sqrt(1.0 * val));
                data_.push_back(val);
            }
            ++it;
        }
    }
}

template <typename T>Image<T>::~Image() {
}

/**
 * Set the entire mask; masked areas are those with '0'
 * @param mask
 */
template <typename T>void Image<T>::setmask(const std::vector<char>& mask) {

    std::vector<char>::const_iterator it;
    std::vector<bool>::iterator itmask;

    itmask = mask_.begin();
    for (it = mask.begin(); it != mask.end(); ++it) {
        if (*it) *itmask = false;
        ++itmask;
    }
}

/**
 * set mask(w,h) to 0 (mask it out)
 * @param w
 * @param h
 */
template <typename T>void Image<T>::setmask(int w, int h) {
    if (w < 0 || h < 0 || w >= width_ || h >= height_) {
        throw std::logic_error("Error: Image::setmask with invalid coordinates.");
    }
    int idx(h * width_ + w);
    mask_[idx] = 0;
}

/**
 * Determine the gradient of the data within the box; no range checking
 * @param cx
 * @param cy
 * @param R
 * @param gx
 * @param gy
 */
template <typename T> void Image<T>::gradient(const int cx, const int cy, const char R, int& gx, int& gy) const {
    gx = 0;
    gy = 0;

    // if it were not for the mask, only the outer rim would contribute,
    // all other values cancel:
    /*
    // x gradient left to right every row
    for (int h = -R; h <= R; ++h) {
        gx += data_[cx+R+(cy+h)*width_];
        gx += data_[cx+R+1+(cy+h)*width_];
        gx -= data_[cx-R-1 + (cy+h)*width_];
        gx -= data_[cx-R + (cy+h)*width_];
    }
    // y-gradient top to bottom every colum
    for (int w = -R; w <= R; ++w) {
        gy += data_[cx+w + (cy+R)*width_];
        gy += data_[cx+w + (cy+R+1)*width_];
        gy -= data_[cx-w + (cy-R-1)*width_];
        gy -= data_[cx-w + (cy-R)*width_];
    }
     */
    // with the mask, every gradient pair must be checked
    for (int h = -R; h <= R; ++h) {
        int idx = cx + (cy + h) * width_;
        for (int w = -R; w <= R; ++w) {
            idx += w;
            if (!(mask_[idx + 1] && mask_[idx - 1])) {
                gx += data_[idx + 1] - data_[idx - 1];
            }
            if (!(mask_[idx + width_] && mask_[idx - width_])) {
                gy += data_[idx + width_] - data_[idx - width_];
            }
        }
    }
}

template <typename T> int Image<T>::sumpixels(const int x0, const int y0, const char R) const{
    int sum(0);

    for (int h = -R; h <= R; ++h) {
        typename std::vector<T>::const_iterator it(data_.begin() + x0 + (y0 + h) * width_);
        std::vector <bool>::const_iterator itmask(mask_.begin() + x0 + (y0 + h) * width_);
        for (int r = -R; r <= R; ++r) {
            if (! *itmask) sum += *it;
            ++it;
            ++itmask;
        }
    }
    return sum;
}

/**
 * Perform the line step up to n steps to return x0 and y0 with maximal sum
 * of pixels in direction of gx/gy
 * @param x0
 * @param y0
 * @param R
 * @param gx
 * @param gy
 * @param pixelsum
 */
template <typename T> void Image<T>::linestep(int& x0, int& y0, const char R,
        const int gx, const int gy, int& pixelsum, const int maxsteps) const {
    int mysum(pixelsum);
    // direction of gradient
    if (gx == 0) {
            T signum (1);
            if (gy < 0)  signum = -1;
            // sign > 0: shift downwards, substract top row
            // sign < 0: shift upwards, substract bottom row
            for (int i = 0; i < maxsteps; ++i) {
                // substract y0 - sign * R
                int idx = (x0 - R) + (y0 - signum*R)*width_;
                for (int r = -R; r <= R; ++r) {
                    if (!mask_[idx]) mysum -= data_[idx];
                    ++idx;
                }
                // shift y0 by sign
                // next row: move up if gradient < 0
                idx = (x0 - R) - (y0 + signum*(R + 1))*width_;
                // next row 
                for (int r = -R; r <= R; ++r) {
                    if (!mask_[idx]) mysum += data_[idx];
                    ++idx;
                }
                // if previous step was better, return w/o updating y0 or pixelsum
                if (mysum <= pixelsum) {
                    return;
                }
                // update y0 and pixelsum
                pixelsum = mysum;
                y0 += signum;
            }
    }
    else if (gy == 0) {
            T signum (1);
            if (gx < 0)  signum = -1;
            // sign > 0: shift right, substract left row
            // sign < 0: shift left, substract right row
            for (int i = 0; i < maxsteps; ++i) {
            int idx = (x0 - signum*R) + (y0 - R)*width_;
            // substract x0 - sign*R
            for (int r = -R; r <= R; ++r) {
                if (!mask_[idx]) mysum = -data_[idx];
                idx += width_;
            }
            // add x0 + sign*(R+1)
            idx = (x0 + signum*(R+1)) + (y0 - R)*width_;
            for (int r = -R; r <= R; ++r) {
                if (!mask_[idx]) mysum = +data_[idx];
                idx += width_;
            }
            if (mysum <= pixelsum) {
                return;
            }
            pixelsum = mysum;
            x0 += signum;
        }
    }
    else { 
        float fgx (1.0 * gx);
        float fgy (1.0 * gy);
        int cx (x0), cy (y0);
        
        const float l (sqrt(fgx*fgx + fgy*fgy));
        fgx /= l;
        fgy /= l;
        for (int i = 1; i < maxsteps; ++i) {
            int x = round (cx + i*fgx);
            int y = round (cy + i*fgy);
            
            mysum = sumpixels(x, y, R);
            if (mysum < pixelsum) {
                return;
            }
            else {
                x0 = x;
                y0 = y;
                pixelsum = mysum;
            }
        }
    }
    
    // if we get here, increase in intensity up to the end
    return;

}

/**
 *  Calculate the intensity weighted centre of the image
 * @param radius side of search window
 * @param x0 starting position. if <0: set to width_/2
 * @param y0 starting position. if < 0: set to height_/2
 * results are stored at x0/y0
 */
template <typename T> void Image<T>::centre(const char radius, int& x0, int& y0) const {
    int gx, gy; // gradient for movement

    // mark already visited fields
    std::vector<bool> visitmap(data_.size(), false);

    // if x0 or y0 are negative, set to peak pixel
    typename std::vector<T>::const_iterator itmax = std::max_element(data_.begin(), data_.end());
    int idx = itmax - data_.begin();
    y0 = idx / width_;
    x0 = idx % width_;
    visitmap [idx] = true;

    // current optimal centre
    int cx(x0), cy(y0);
    // mark as visited

    int bestsum = sumpixels(x0, y0, radius);
    // mark x0, y0 as visited
    do {
        gradient(x0, y0, radius, gx, gy);
        if (gx == 0 && gy == 0) {
            // update centre and exit
            break;
        }
        linestep(x0, y0, radius, gx, gy, bestsum, 100);
        int idx = x0 + y0*width_;
        if (visitmap [idx]) {
            break;
        }
        cx = x0;
        cy = y0;

    } while (!(gx == 0 && gy == 0));
    
    std::cout << "---> Centre of image based on " << int(radius) << " x " 
            << int(radius) << " box: (" << x0 << ", " << y0 << ")\n";
    x0 = cx;
    y0 = cy;
}

template <typename T> void Image <T>::add (Image<int>& image) const {
    for (size_t i = 0; i < this->width()*this->height(); ++i) {
        image.data(i) += this->data(i);
    }
}

// template class Image<unsigned char>;
template class Image<unsigned short>;
template class Image<int>;
template class Image<unsigned int>;
