/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */

#include "io.h"
#include "myExceptions.h"
#include "defines.h"

#include <tiffio.h>
#include <cbf/cbf.h>
#include <iostream>


/**
 * Read 2D data from a TIFF file. Data are assumed to fit into unsigned short,
 * an error message is printed if BitsPerPixel > 16
 * @param filename
 * @param data
 * @param width
 * @param height
 * @param defaultbpp Unless 0 check the bitsPerPixel with this value
 * @return 
 */
int IO::readtiff(const char* filename,
        std::vector<unsigned int>& data,
        int& width,
        int& height) {
    TIFF* tiff;
    int ret;
    uint16 bitsPerPixel;
    uint16 samplesPerPixel;

    TIFFSetWarningHandler(NULL);

    tiff = TIFFOpen(filename, "r");
    if (tiff == NULL) {
        printf("*** Error: Could not open \"%s\". Exiting.\n", filename);
        return (-1);
    }

    ret = TIFFGetField(tiff, TIFFTAG_IMAGEWIDTH, &width);
    if (ret != 1) {
        printf("File Format Error: 'width' not declared in file header. Exiting\n");
        return -1;
    }
    ret = TIFFGetField(tiff, TIFFTAG_IMAGELENGTH, &height);
    if (ret != 1) {
        printf("File Format Error: 'height' not declared in file header. Exiting\n");
        return -1;
    }
    ret = TIFFGetField(tiff, TIFFTAG_BITSPERSAMPLE, &bitsPerPixel);
    if (ret != 1) {
        printf("File Format Error: 'bitsPerPixel' not declared in file header. Exiting\n");
        return -1;
    }
    
    ret = TIFFGetField(tiff, TIFFTAG_SAMPLESPERPIXEL, &samplesPerPixel);
    if (ret != 1) {
        samplesPerPixel = 1;
    }

    printf("# Opening TIFF image \"%s\" %u x %u, %u bits per pixel\n",
            filename, width, height, bitsPerPixel);


    // reading data in strips; adapted from http://www.remotesensing.org/libtiff/libtiff.html#strips
    tsize_t scanlinelength = TIFFScanlineSize(tiff); 
    tdata_t tiffstrip = _TIFFmalloc(scanlinelength);
    data.resize(width*height, 0);

    switch (bitsPerPixel) {
        case 8:
            for (int row = 0; row < height; ++row) {
                TIFFReadScanline(tiff, tiffstrip, row);
                for (int i = 0; i < width; ++i) {
                    data[row * width + i] = ((unsigned char*) tiffstrip)[i];
                    std::cout << "---> Copied data of value " << data[row * width + i]
                            << std::endl;
                }
            }
            break;
        case 16:
            for (int row = 0; row < height; ++row) {
                TIFFReadScanline(tiff, tiffstrip, row);
                for (int i = 0; i < width; ++i) {
                    data[row * width + i] = ((uint16*) tiffstrip)[i];
                }
            }
            break;
        case 32:
            for (int row = 0; row < height; ++row) {
                TIFFReadScanline(tiff, tiffstrip, row);
                for (int i = 0; i < width; ++i) {
                    data[row * width + i] = ((uint32*) tiffstrip)[i];
                }
            }
            break;
        default:
        printf ("Error: Only TIFF images with 8 and 16 bits per pixel supported. Exiting\n");
        throw myExcepts::ImageFormat(filename);
            
    }
    
    _TIFFfree(tiffstrip);
    TIFFClose(tiff);

    return bitsPerPixel;
}

void IO::cbf(const char* filename, const Image<unsigned int>& image) {
    FILE* outp;
    outp = std::fopen(filename, "w");

    

    cbf_handle cbf;
    cbf_make_handle(&cbf);
    // create a CBF
    cbf_new_datablock(cbf, filename);

    // put a miniCBF header
    cbf_new_category(cbf, "array_data");
    cbf_new_column(cbf, "header_convention");
    cbf_set_value(cbf, "PILATUS_1.2");

    cbf_new_column(cbf, "header_contents");
    cbf_set_value(cbf, Defs::Timepix::cbf_header);

    // put the image
    cbf_new_category(cbf, "array_data");
    cbf_new_column(cbf, "data");
    std::cout << "---> Creating image of size " << image.width() * image.height() 
            << " pixel" << std::endl;
    unsigned int* image_data;
    image_data = new unsigned int  [image.width() * image.height()] ;
    
    //    int image_data [image.width() * image.height()];
    for (size_t i = 0; i < image.width() * image.height(); ++i) {
        image_data[i] = image.data(i);
    }
    cbf_set_integerarray_wdims_fs(cbf, CBF_BYTE_OFFSET, 1, image_data, sizeof (image_data[0]),
            1, image.width() * image.height(), "little_endian", image.width(),
            image.height(), 0, 1);

    cbf_write_widefile(cbf, outp, 1, CBF, MSG_DIGEST | MIME_HEADERS | PAD_4K, 0);
    //    fclose (outp);
    // no need to fclose() here as the 3rd argument "readable" is 1
    cbf_free_handle(cbf);
    delete [] image_data;
}
